from .exception import ParameterError
from .nvs import Nvs


class Decoder:
    """Class responsible for decoding binary NVS."""

    def __init__(self, sector_size: int):
        """Crate nvs decoder.

        :param int sector_size: Size of single NVS sector
        """
        self.sector_size = sector_size

    def load(self, data: bytes) -> dict[int, bytes]:
        """
        Loads NVS from bytes.
        
        :param bytes data: NVS data.
        """
        if not len(data) % self.sector_size == 0:
            raise ParameterError("NVS size must be a multiplicity of sector_size.")

        entries: dict[int, bytes] = {}

        for sector in Nvs(data, self.sector_size):
            for ate in sector:
                entry = ate.get_entry()
                entries[entry.id] = entry.value


        return {id: data for id, data in entries.items() if data is not None}




from typing import cast
from .ate import Ate
from .exception import ChecksumError, EncodingError, DecodingError
from .entry import Entry


class Sector:
    """Class responsible for decoding and iterating over NVS sector."""

    def __init__(self, data: bytes):
        """
        Creates Sector.
        
        :param bytes data: Sector data
        """
        self.data = data
        self.sector_valid = True

        try:
            # GC done ate
            gc_done = Ate.from_bytes(len(self.data) - Ate._SIZE * 2, self.data)
            if gc_done is None or not gc_done.is_gc_done:
                self.sector_valid = False
        except ChecksumError as _:
            self.sector_valid = False

    @property
    def is_closed(self):
        try:
            return Ate.from_bytes(len(self.data) - Ate._SIZE, self.data).is_close
        except:
            return False

    def __iter__(self):
        self.next_ate_offset = len(self.data) - Ate._SIZE * 3  # skip close_ate and gc_done ate
        self.last_ate_offset = cast(Ate, Ate.from_bytes(len(self.data) - Ate._SIZE, self.data)).data_offset if self.is_closed else None
        self.data_offset = 0
        return self

    def __next__(self) -> Ate:
        if not self.sector_valid:
            raise StopIteration

        # If sector is closed and we already processed last ate, do not even try to decode next ate
        if self.last_ate_offset is not None and self.next_ate_offset < self.last_ate_offset:
            raise StopIteration

        # Data reached allocation table, sector should be closed
        if self.next_ate_offset < self.data_offset:
            raise StopIteration

        ate = Ate.from_bytes(self.next_ate_offset, self.data)

        if ate:
            self.data_offset = ate.data_offset + ate.aligned_data_size
            if self.data_offset > self.next_ate_offset:
                raise DecodingError("Data overlaps with allocation table")

            self.next_ate_offset -= Ate._SIZE
            return ate
        else:
            raise StopIteration


class SectorBuilder:
    """Helper class for building NVS sector."""

    def __init__(self, sector_size: int):
        """
        Creates SectorBuilder.

        :param int sector_size: Size of sector
        """
        self.sector_size = sector_size
        self.data = bytearray.fromhex("FF") * self.sector_size

        # gc done ate
        self.last_ate = Ate(self.sector_size - Ate._SIZE * 2, 0xFFFF, None, 0)
        self.last_ate.to_bytes(self.data)

    def add(self, entry: Entry) -> bool:
        """
        Adds entry to sector.
        
        :param Entry entry: Entry to add
        :return: True if entry added successfully, False otherwise
        """
        ate = self.last_ate.next(entry.id, entry.value)
        try:
            ate.to_bytes(self.data)
            self.last_ate = ate

            # Automatically close sector is there is less space than for single ate
            if ate.ate_offset - (ate.data_offset + ate.aligned_data_size) < Ate._SIZE:
                self.close()

            return True
        except EncodingError as _:
            return False

    def close(self):
        """
        Marks sector as closed.
        """
        Ate(self.sector_size - Ate._SIZE, 0xFFFF, None, self.last_ate.ate_offset).to_bytes(self.data)

    def get(self) -> Sector:
        """
        Returns built Sector.

        :return: Sector
        """
        return Sector(self.data)

    def get_bytes(self) -> bytes:
        """
        Returns bytes of built Sector.

        :return: Sector bytes
        """
        return bytes(self.data)



from itertools import islice


def batched(iterable, n):
    # batched('ABCDEFG', 3) --> ABC DEF G
    if n < 1:
        raise ValueError('n must be at least one')
    it = iter(iterable)
    while batch := tuple(islice(it, n)):
        yield batch




from .sector import Sector


class Nvs:
    """Class responsible for iterating over NVS."""

    def __init__(self, data: bytes, sector_size: int):
        """
        Constructs NVS.
        
        :param bytes data: NVS data
        :param int sector_size: Size of single NVS sector
        """
        self.sectors = [Sector(data[i:i + sector_size]) for i in range(0, len(data), sector_size)]
        first_idx = None
        last_idx = None
        for idx in range(len(self.sectors)):
            if self.sectors[idx].is_closed:
                if last_idx is not None:
                    first_idx = idx
                    break
                elif first_idx is None:
                    first_idx = idx
            else:
                if last_idx is None:
                    last_idx = idx

        # sort sectors
        self.sectors = self.sectors[first_idx:] + self.sectors[:first_idx]

    def __iter__(self):
        self.idx = 0
        return self

    def __next__(self) -> Sector:
        if self.idx < len(self.sectors):
            self.idx += 1
            return self.sectors[self.idx - 1]
        raise StopIteration



class ZnvsException(Exception):
    """Base for all znvs exceptions."""
    pass


class ParameterError(ZnvsException):
    """Error raised when input parameters validation fails."""
    pass


class ChecksumError(ZnvsException):
    """Error raised when NVS entry validation fails."""
    pass


class DecodingError(ZnvsException):
    """NVS decoding error"""
    pass

class EncodingError(ZnvsException):
    """NVS encoding error"""
    pass




from __future__ import annotations
import math
import struct
from crc import Calculator, Configuration
from .entry import Entry
from .exception import ChecksumError, EncodingError, ParameterError


class Ate:
    """Class responsible for Allocation Table Entry decoding/encoding."""

    _SIZE = 8
    _DATA_ALIGNMENT = 4
    _CRC_CALCULATOR = Calculator(Configuration(width=8, polynomial=0x07, init_value=0xff))

    def __init__(self, ate_offset: int, data_id: int, data: bytes, data_offset: int):
        """
        Initializes ATE.

        :param int ate_offset: Offset of ATE
        :param int data_id: ID of data to be allocated
        :param bytes data: Data to be allocated
        :param int data_offset: Offset of data in sector
        """
        if data_offset % Ate._DATA_ALIGNMENT or ate_offset % Ate._DATA_ALIGNMENT:
            raise ParameterError("Offset not aligned")

        self.ate_offset = ate_offset
        self.data_id = data_id
        self.data_offset = data_offset
        self.data = data

    def get_entry(self) -> Entry | None:
        """
        Returns Entry if ATE is not special kind (close of gc).

        :return: Entry or None
        """
        if self.is_close or self.is_gc_done:
            return None
        return Entry(self.data_id, self.data)

    @property
    def data_len(self):
        return 0 if self.data is None else len(self.data)

    @property
    def is_close(self):
        return self.data_id == 0xFFFF and self.data is None and self.data_offset != 0x00

    @property
    def is_gc_done(self):
        return self.data_id == 0xFFFF and self.data is None and self.data_offset == 0x00

    @property
    def aligned_data_size(self):
        return Ate._DATA_ALIGNMENT * math.ceil(self.data_len/Ate._DATA_ALIGNMENT)

    def to_bytes(self, sector_data: bytearray):
        """
        Serializes ATE into given sector.

        :param bytearray sector_data: NVS sector data
        """
        if self.data_offset + self.aligned_data_size > self.ate_offset:
            raise EncodingError("Data does not fit into sector")

        if sector_data[self.ate_offset:self.ate_offset + Ate._SIZE] != b'\xFF' * Ate._SIZE or \
                sector_data[self.data_offset:self.data_offset + self.aligned_data_size] != b'\xFF' * self.aligned_data_size:
            raise EncodingError("Memory not empty")

        ate = struct.pack("<HHHB", self.data_id, self.data_offset, self.data_len, 0xFF)
        ate += Ate._calc_crc(ate).to_bytes(1, 'little')
        if self.data_len:
            sector_data[self.data_offset:self.data_offset + self.data_len] = self.data
        sector_data[self.ate_offset:self.ate_offset + Ate._SIZE] = ate

    def next(self, data_id: int, data: bytes) -> Ate:
        """
        Creates next Ate with given data_id and data.

        :param int data_id: ID of data to be allocated
        :param bytes data: Data to be allocated
        """
        return Ate(self.ate_offset - Ate._SIZE, data_id, data, self.data_offset + self.aligned_data_size)

    @staticmethod
    def _calc_crc(allocation_table_entry: bytes) -> int:
        return Ate._CRC_CALCULATOR.checksum(allocation_table_entry)

    @staticmethod
    def _validate_crc(allocation_table_entry: bytes) -> bool:
        return 0 == Ate._calc_crc(allocation_table_entry)

    @staticmethod
    def from_bytes(ate_offset: int, sector_data: bytes) -> Ate | None:
        """
        Deserializes Ate from bytes.

        :param int ate_offset: Offset of Ate
        :param bytearray sector_data: NVS sector data
        :return: ATE or None if ate_offset points to empty sector data
        """
        ate_data = sector_data[ate_offset:ate_offset + Ate._SIZE]
        if ate_data == bytes.fromhex("FFFFFFFFFFFFFFFF"):
            return None

        # Each entry is 8 bytes
        # 0 0 1 1 2 2 3 3 4 4 5 5 6 6 7 7
        # |  ID  | OFFS  |  LEN  | - |CRC|
        data_id, data_offset, data_length = struct.unpack_from("<HHH", ate_data)
        if not Ate._validate_crc(ate_data):
            raise ChecksumError(f"ATE CRC is invalid. Data id={hex(data_id)}")
        return Ate(ate_offset, data_id, sector_data[data_offset:data_offset + data_length] if data_length > 0 else None, data_offset)



from .exception import EncodingError
from .entry import Entry
from .sector import SectorBuilder
import logging

logger = logging.getLogger(__name__)


class Encoder:
    """Class responsible for encoding data to binary NVS image."""

    def __init__(self, sector_size: int):
        """Crate nvs encoder.

        Arguments:
        :param int sector_size: Size of single NVS sector
        """
        self.sector_size = sector_size

    def dump(self, entries: dict[int, bytes], force_close: bool = False) -> bytes:
        """
        Serializes NVS entries to bytes.
        
        :param list[Entry] entries: List of Entries to be put into NVS
        :param bool force_close: Forcibly closes every sector, even if not full
        :return: NVS with serialized Entries
        :raises EncodingError: if Entry cannot be fit into sector
        """
        nvs = b''
        while len(entries) > 0:
            sector = SectorBuilder(self.sector_size)
            consumed: list[int] = []
            for id, value in entries.items():
                if sector.add(Entry(id, value)):
                    consumed.append(id)
                else:
                    # If sector is empty and entry could not be added it means it cannot fit into any sector
                    if sum(1 for _ in sector.get()) == 0:
                        raise EncodingError(f"Entry {id=} cannot fit into sector of {self.sector_size=}")
                    sector.close()
                    break

            # Remove entries that were successfully encoded
            for id in consumed:
                entries.pop(id)

            if force_close:
                sector.close()

            nvs = nvs + sector.get_bytes()

        return nvs
